/*---------------------------*/
/*         Shadow.fx         */
/*        -----------        */
/*---------------------------*/
//  Shaders and techniques   //
//  to generate and render   //
//  shadow maps.             //
//---------------------------//

#define SHADOW_EPSILON 0.0001f
#define AMBIENT_AMOUNT 0.3f
#define SHADOW_AMOUNT  0.5f


// Extern variables //

// Matrices //
matrix  lightViewMatrix;
matrix  lightProjMatrix;
matrix  camViewMatrix;
matrix  camProjMatrix;
matrix  worldMatrix;
matrix  texMatrix;

// Texturas //
texture normalTex;
texture shadowMap;

// Material //
float4  matColor;

// Luz //
vector  lightVector;
float4  lightColor;

// Tamao //
int     mapSize;

// Texture Sampler //
sampler2D texSampler =
sampler_state
{
	Texture   = <normalTex>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Linear;
};

// Shadow Map Sampler //
sampler2D mapSampler =
sampler_state
{
	Texture   = <shadowMap>;
	MinFilter = Point;
	MagFilter = Point;
	MipFilter = Point;
	AddressU  = Clamp;
	AddressV  = Clamp;
};


//--------------------------------------
// Auxiliar Function: Calculate Lighting
//--------------------------------------
float Lighting( in vector inPos,
      		in float3 normal )
{
	float3 lightDir;
	if( lightVector.w == 0.0f )
	    lightDir = normalize( -lightVector.xyz );
	else
	    lightDir = normalize( lightVector.xyz - inPos.xyz );

	return saturate( saturate( dot( normal, lightDir ) ) + AMBIENT_AMOUNT );
}


//-----------------------------------
// Vertex Shader: Shadow Map Creation
//-----------------------------------
void mapVertex( in  vector inPos    : POSITION,
     		out vector outPos   : POSITION,
		out float2 outDepth : TEXCOORD0 )
{
	// Vrtice en espacio de la luz
	outPos = mul( inPos , worldMatrix );
	outPos = mul( outPos, lightViewMatrix );
	outPos = mul( outPos, lightProjMatrix );
	
	outDepth.xy = outPos.zw;
}


//----------------------------------
// Pixel Shader: Shadow Map Creation
//----------------------------------
void mapPixel( in  float2 inDepth  : TEXCOORD0,
     	       out float4 outColor : COLOR )
{
	// Profundidad para guardar en el surface
	outColor = inDepth.x / inDepth.y;
}


//------------------------------------
// Vertex Shader: Shadow Map Rendering
//------------------------------------
void shadowVertex( in  vector inPos        : POSITION,
     		   in  float3 inNormal     : NORMAL,
     		   in  float2 inTexCoords  : TEXCOORD0,
     		   out vector outPos       : POSITION,
		   out float2 outTexCoords : TEXCOORD0,
		   out float4 outMapCoords : TEXCOORD1,
		   out vector outPosWorld  : TEXCOORD2,
		   out float3 outNormal    : TEXCOORD3 )

{
	// Vrtice en espacio del mundo
	outPosWorld = mul( inPos, worldMatrix );
	// Normal en espacio del mundo
	outNormal   = mul( inNormal, (float3x3)worldMatrix );

	// Vrtice en espacio de la cmara
	outPos = mul( outPosWorld, camViewMatrix );
	outPos = mul( outPos, camProjMatrix );

	// Replico las coordenadas de textura
	outTexCoords = inTexCoords;

	// Transformacin del vrtice a coordendadas de textura
	outMapCoords = mul( outPosWorld, texMatrix );
}


//-----------------------------------
// Pixel Shader: Shadow Map Rendering
//-----------------------------------
void shadowPixel( in  float2 inTexCoords : TEXCOORD0,
     		  in  float4 inMapCoords : TEXCOORD1,
		  in  vector inPosWorld  : TEXCOORD2,
		  in  float3 inNormal    : TEXCOORD3,
     		  out float4 outColor    : COLOR )
{
	// Intensidad de la sombra
	float shadow;
	
	// Clculo de luz
	float light = Lighting( inPosWorld, inNormal );

	// Color de la textura
	vector texColor = tex2D( texSampler, inTexCoords );

	// Profundidad en el shadow map
	float depth = tex2Dproj( mapSampler, inMapCoords ) + SHADOW_EPSILON;

	// Si es menor la profundiad en el shadow map el pixel est en sombra
	if( depth < ( inMapCoords.z / inMapCoords.w ) )
	    shadow = SHADOW_AMOUNT;
	else
	    shadow = 1.0f;

	outColor   = matColor * lightColor * texColor * light * shadow;
	outColor.a = 1.0f;
}


//-----------------------------
// Technique: Create Shadow Map
//-----------------------------
technique CreateShadowMap
{
	pass p0
	{
		VertexShader = compile vs_2_0 mapVertex();
		PixelShader  = compile ps_2_0 mapPixel();
	}
}


//-----------------------------
// Technique: Render Shadow Map
//-----------------------------
technique RenderShadowMap
{
	pass p0
	{
		VertexShader = compile vs_2_0 shadowVertex();
		PixelShader  = compile ps_2_0 shadowPixel();
	}
}